from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import json
import os

from confusionflow.logging.utils import (
    check_folderpath,
    update_runindex,
    update_datasetindex,
    create_logdir,
    create_dataset_config,
)
from confusionflow.logging import FoldLog


class Run:
    """
    Run is a simple wrapper for simplifying the logging of an experiment.
    """

    def __init__(self, runId, folds, trainfoldId):
        self.runId = runId
        self.folds = folds
        self.trainfoldId = trainfoldId
        self.foldlogs = list()

        for fold in self.folds:
            foldlog = self.create_foldlog(fold.foldId)
            self.foldlogs.append(foldlog)

    def get_keras_callback(self, loss):
        from confusionflow.logging.callbacks import RunLogger

        runlogger = RunLogger(self, loss)
        return runlogger

    def export(self, logdir):
        create_logdir(logdir)
        run_path = check_folderpath(os.path.join(logdir, "runs"))
        filepath = os.path.join(run_path, self.runId + ".json")
        with open(filepath, "w") as outfile:
            json.dump(self.asdict(), outfile)

        update_runindex(logdir)

        for fold in self.folds:
            create_dataset_config(logdir, fold.dataset_config)
        update_datasetindex(logdir)

        for foldlog in self.foldlogs:
            foldlog.export(logdir)

    def create_foldlog(self, foldId):
        foldlogId = self.runId + "_" + foldId
        return FoldLog(foldlogId, self.runId, foldId)

    def asdict(self):
        d = dict()
        d["runId"] = self.runId
        d["trainfoldId"] = self.trainfoldId
        d["hyperparam"] = dict()
        d["foldlogs"] = []
        for foldlog in self.foldlogs:
            d["foldlogs"].append(foldlog.asdict())

        return d



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import os


def check_folderpath(folderpath):
    """Checks whether supplied folderpath is a path to a directory"""
    folderpath = os.path.realpath(folderpath)
    if os.path.isdir(folderpath):
        return folderpath
    else:
        raise OSError("Error! `{}` is not a path to a valid folder".format(folderpath))


def get_logdir_from_env():
    """Tries to set the <path_to_logdir> from the CONFUSIONFLOW_LOGDIR environment variable"""
    logdir = os.environ.get("CONFUSIONFLOW_LOGDIR", None)
    if logdir is None:
        raise OSError(
            "Error! Please specify folder with the logs via `export CONFUSIONFLOW_LOGDIR=<path to logdir>`"
        )
    else:
        return logdir



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import json
import os

from confusionflow.logging.utils import check_folderpath
from confusionflow.logging.foldlogdata import FoldLogData


class FoldLog:
    """
    A FoldLog is a performance log of a model for a fold.
    """

    def __init__(self, foldlogId, runId, foldId):
        self.foldlogId = foldlogId
        self.description = ""
        self.runId = runId
        self.foldId = foldId
        self.foldlogdata = FoldLogData(foldlogId)

    def add_epochdata(self, epochId, confmat):
        self.foldlogdata.add_epochdata(epochId, confmat)

    def export(self, logdir):
        foldlog_path = check_folderpath(os.path.join(logdir, "foldlogs"))
        filepath = os.path.join(foldlog_path, self.foldlogId + ".json")
        with open(filepath, "w") as outfile:
            json.dump(self.asdict(), outfile)

        self.foldlogdata.export(logdir)

    def asdict(self):
        return {
            "foldlogId": self.foldlogId,
            "description": self.description,
            "runId": self.runId,
            "foldId": self.foldId,
            "foldlogdataId": self.foldlogdata.get_id() + "_data",
            "numepochs": self.foldlogdata.get_numepochs(),
        }



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import argparse

from gevent import pywsgi

from confusionflow.server import create_app


def main():
    parser = argparse.ArgumentParser(description="ConfusionFlow CLI")
    parser.add_argument("--host", type=str, default="localhost")
    parser.add_argument("--logdir", type=str, required=True)
    parser.add_argument("--port", type=int, default=8080)

    FLAGS = parser.parse_args()

    confusionflow_app = create_app(FLAGS.logdir)

    http_server = pywsgi.WSGIServer(
        (FLAGS.host, FLAGS.port), confusionflow_app)

    try:
        print("Starting ConfusionFlow Server on http://{}:{}".format(
            FLAGS.host, FLAGS.port))
        http_server.serve_forever()
    except KeyboardInterrupt:
        print("Server received KeyboardInterrupt. Shutting down ...")


if __name__ == "__main__":
    main()



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import json
import os

from confusionflow.logging.utils import check_folderpath


class FoldLogData:
    """
    A FoldLogData is contains the EpochData for a specific FoldLog.
    """

    def __init__(self, foldlogId):
        self.foldlogId = foldlogId
        self.numepochs = 0
        self.epochdata = list()

    def add_epochdata(self, epochId, confmat):
        epochdata = dict()
        epochdata["epochId"] = epochId
        epochdata["confmat"] = confmat

        self.epochdata.append(epochdata)
        self.numepochs = len(self.epochdata)

    def get_id(self):
        return self.foldlogId

    def get_numepochs(self):
        return self.numepochs

    def export(self, logdir):
        foldlogdata_path = check_folderpath(os.path.join(logdir, "foldlogdata"))
        filepath = os.path.join(foldlogdata_path, self.foldlogId + "_data.json")
        with open(filepath, "w") as outfile:
            json.dump(self.asdict(), outfile)

    def asdict(self):
        d = dict()
        d["foldlogId"] = self.foldlogId
        d["numepochs"] = self.numepochs
        d["epochdata"] = self.epochdata

        return d



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import json
import os

from confusionflow.logging.utils import check_folderpath


class Fold:
    """
    A Fold is a subset of your dataset
    """

    def __init__(self, dataset, foldId, dataset_config):
        self.dataset = dataset
        self.foldId = foldId
        self.description = ""
        self.dataset_config = dataset_config



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from flask import Blueprint, redirect

from confusionflow.server.utils import serve_file


static_file_path = None

bp = Blueprint("static", __name__)


@bp.record
def record_satic_file_path(setup_state):
    """Sets the static_file_path based on the config during Blueprint setup."""
    global static_file_path

    config = setup_state.app.config
    static_file_path = config["STATIC_FILE_PATH"]


@bp.route("/")
def index():
    """Returns 'index.html' of 'ui' component"""
    return serve_file(static_file_path, "index.html", "")


@bp.route("/<path:filename>")
def serve_static(filename):
    """Serves files from 'static_file_folder'."""
    return serve_file(static_file_path, filename, "")



import torch
import torchnet as tnt


def log_epoch(run, model, device, epoch, numclass):
    for fold, foldlog in zip(run.folds, run.foldlogs):
        assert fold.foldId == foldlog.foldId

        data_loader = fold.dataset
        log_performance(foldlog, model, device, data_loader, epoch, numclass)


def log_performance(foldlog, model, device, data_loader, epoch, numclass):
    model.eval()

    confusion_matrix = tnt.meter.ConfusionMeter(numclass)

    with torch.no_grad():
        for data, target in data_loader:
            data, target = data.to(device), target.to(device)
            output = model(data)
            pred = output.max(1, keepdim=True)[1]

            confusion_matrix.add(pred.data.squeeze(), target.long())

    foldlog.add_epochdata(epochId=epoch, confmat = confusion_matrix.conf.flatten().tolist())




from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import os

from flask import send_from_directory


def serve_file(foldername, filename, errormsg):
    if not os.path.isfile(os.path.join(foldername, filename)):
        return errormsg
    else:
        return send_from_directory(foldername, filename)



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import os

from flask import Flask

import confusionflow
from confusionflow.server.blueprints import api, web
from confusionflow.utils import check_folderpath, get_logdir_from_env


def create_app(logdir=None):
    app = Flask(__name__)

    if logdir is None:
        logdir = get_logdir_from_env()

    static_file_path = os.path.join(
        os.path.dirname(os.path.realpath(confusionflow.__file__)), "static"
    )

    # setup api
    app.config["LOGDIR"] = check_folderpath(logdir)
    app.register_blueprint(api.bp, url_prefix="/api")

    # setup web
    # only setup web if not in development mode
    if app.config["ENV"] != "development":
        app.config["STATIC_FILE_PATH"] = check_folderpath(static_file_path)
        app.register_blueprint(web.bp)
        app.add_url_rule("/", endpoint="index")

    return app



import setuptools.command.build_py
import subprocess
import os

from setuptools import Command
from setuptools import find_packages, setup


class build_ui(setuptools.Command):
    user_options = []

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        subprocess.call(["npm", "install"], cwd="./confusionflow/ui")
        subprocess.call(["npm", "run", "build"], cwd="./confusionflow/ui")
        subprocess.call(
            ["cp", "-r", "./confusionflow/ui/build", "./confusionflow/static"])


class build_py(setuptools.command.build_py.build_py):
    def run(self):
        if not os.path.exists("./confusionflow/static"):
            self.run_command('build_ui')
        setuptools.command.build_py.build_py.run(self)


here = os.path.abspath(os.path.dirname(__file__))
with open(os.path.join(here, 'confusionflow/VERSION')) as version_file:
    version = version_file.read().strip()


readme = open('README.md').read()


requirements = [
    "flask",
    "gevent",
    "pyyaml"
]

cmdclass = {
    'build_py': build_py,
    'build_ui': build_ui,
}

entry_points = {
    "console_scripts": [
        "confusionflow=confusionflow:main"
    ]
}

setup(
    # Metadata
    name="confusionflow",
    version=version,
    author="ConfusionFlow Contributors",
    author_email="gfrogat@gmail.com",
    url="https://github.com/confusionflow/confusionflow",
    license="BSD-3-Clause",
    description="ConfusionFlow is a visualization tool that enables more nuanced monitoring of a neural network's training process.",
    long_description=readme,
    long_description_content_type='text/markdown',
    project_urls={
        'Documentation': 'https://docs.confusionflow.org',
        'Source': 'https://github.com/confusionflow/confusionflow/',
        'Tracker': 'https://github.com/confusionflow/confusionflow/issues',
    },

    # Package info
    cmdclass=cmdclass,
    packages=find_packages(),
    package_data={"confusionflow": ["static/*.*"]},
    include_package_data=True,
    zip_safe=False,
    install_requires=requirements,
    entry_points=entry_points,
)



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import os

from flask import Blueprint, redirect

from confusionflow.server.utils import serve_file


logdir = None

bp = Blueprint("api", __name__)


@bp.record
def record_logdir(setup_state):
    """Sets the logdir based on the config during Blueprint setup."""
    global logdir

    config = setup_state.app.config
    logdir = config["LOGDIR"]


@bp.route("/")
def api_hello():
    return "Welcome to ConfusionFlow API"


@bp.route("/runs")
def get_runs():
    """Returns a list of all available runs as a JSON file."""
    runfolder = os.path.join(logdir, "runs")
    return serve_file(runfolder, "index.json", "Could not load runs.")


@bp.route("/run/<runId>")
def get_run_by_id(runId):
    """Returns the run <runId> as a JSON file."""
    runfolder = os.path.join(logdir, "runs")
    filename = runId + ".json"
    return serve_file(runfolder, filename, "runId not found")


@bp.route("/foldlog/<foldlogId>")
def get_foldlog_by_id(foldlogId):
    """Returns the foldlog <foldlogId> as a JSON file."""
    foldlogfolder = os.path.join(logdir, "foldlogs")
    filename = foldlogId + ".json"
    return serve_file(foldlogfolder, filename, "foldlogId not found")


@bp.route("/foldlog/<foldlogId>/data")
def get_foldlogdata_by_id(foldlogId):
    """Returns the data for foldlog <foldlogId> as a JSON file."""
    foldlogdatafolder = os.path.join(logdir, "foldlogdata")
    filename = foldlogId + "_data.json"
    return serve_file(foldlogdatafolder, filename, "data for foldlogId not found")


@bp.route("/datasets")
def get_datasets():
    """Returns a list of all avaliable datasets as a JSON file."""
    datasetsfolder = os.path.join(logdir, "datasets")
    return serve_file(datasetsfolder, "index.json", "Could not load datasets.")


@bp.route("/dataset/<datasetId>")
def get_dataset_by_id(datasetId):
    """Returns the dataset <datasetId> as a JSON file."""
    datasetsfolder = os.path.join(logdir, "datasets")
    filename = datasetId + ".json"
    return serve_file(datasetsfolder, filename, "datasetId not found.")


@bp.route("/views")
def get_views():
    return "not implemented yet"


@bp.route("/view/<viewId>")
def get_view_by_id(viewId):
    return "not implemented yet"



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import json
import os

import yaml


def create_logdir(logdir):
    logdir = os.path.realpath(logdir)
    if not os.path.exists(logdir):
        os.mkdir(logdir)
        for subfolder in ["/datasets", "/foldlogdata", "/foldlogs", "/runs", "/views"]:
            os.mkdir(logdir + subfolder)


def create_dataset_config(logdir, template_file_path):
    if not os.path.exists(os.path.realpath(template_file_path)):
        raise OSError(
            "Error! File `{}` not found!".format(
                os.path.realpath(template_file_path))
        )

    with open(template_file_path) as f:
        content = f.read()
        data = yaml.load(content)

    export_dict = dict()
    export_dict["datasetId"] = data["dataset"]
    export_dict["description"] = data["description"]
    export_dict["numclass"] = len(data["classes"])
    export_dict["numfolds"] = len(data["folds"])
    export_dict["classes"] = list(map(str, data["classes"]))
    export_dict["folds"] = []

    for foldentry in data["folds"]:
        foldname, folddata = list(foldentry.items())[0]
        export_fold = dict()
        export_fold["foldId"] = "{}_{}".format(
            export_dict["datasetId"], foldname)
        export_fold["description"] = folddata["description"]
        export_fold["dataset"] = export_dict["datasetId"]
        export_fold["numinstances"] = 0
        export_fold["classcounts"] = []

        for classfrequency in folddata["classfrequencies"]:
            classname, instancecount = list(classfrequency.items())[0]
            export_classcount = dict()
            export_classcount["classname"] = str(classname)
            export_classcount["instancecount"] = instancecount
            export_fold["numinstances"] += instancecount
            export_fold["classcounts"].append(export_classcount)

        export_dict["folds"].append(export_fold)

    datasetfolder = logdir + "/datasets/"
    filename = data["dataset"] + ".json"
    with open(datasetfolder + filename, "w") as outfile:
        outfile.write(json.dumps(export_dict))

    update_datasetindex(logdir)


def check_folderpath(folderpath):
    folderpath = os.path.realpath(folderpath)
    if os.path.isdir(folderpath):
        return folderpath
    else:
        raise OSError("Error! Please specify a valid folder {}".format(folderpath))


def update_runindex(logdir):
    runfolder = logdir + "/runs/"
    runfiles = os.listdir(runfolder)

    update_index(runfolder, runfiles)


def update_datasetindex(logdir):
    datasetfolder = logdir + "/datasets/"
    datasetfiles = os.listdir(datasetfolder)

    update_index(datasetfolder, datasetfiles)


def update_index(folder, files):
    index = []
    for file in files:
        if file != "index.json":
            with open(folder + file, "r") as f:
                index.append(json.loads(f.read()))

    with open(folder + "index.json", "w") as f:
        f.write(json.dumps(index))



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import numpy as np
import tensorflow as tf
from tensorflow.keras.callbacks import Callback
from tensorflow import confusion_matrix, Session


class RunLogger(Callback):
    """
    A Runlogger is a simple wrapper for logging confusion matrices when using Keras
    """

    def __init__(self, run, loss):
        self.run = run
        self.loss = loss
        self.session = tf.Session()

    def on_epoch_begin(self, epoch, logs={}):
        for fold, foldlog in zip(self.run.folds, self.run.foldlogs):
            assert fold.foldId == foldlog.foldId

            self.log_performance(fold, foldlog, epoch)

    def log_performance(self, fold, foldlog, epoch):
        x, y = fold.dataset
        predictions = np.argmax(self.model.predict(x, verbose=0), axis=1)
        if self.loss == "categorical_crossentropy":
            targets = np.argmax(y, axis=1)
        elif self.loss == "sparse_categorical_crossentropy":
            targets = y
        else:
            raise ValueError("loss `{}` is not supported".format(self.loss))
        confmat = confusion_matrix(targets, predictions)
        # transform tf.Tensor to list
        confmat = confmat.eval(session=self.session).flatten().tolist()
        foldlog.add_epochdata(epochId=epoch, confmat=confmat)
